/* Copyright 2012, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED,
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED,
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL,
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */
/*
 * EVGPIO - event driven GPIO core
 *
 * The basic idea of EVGPIO cores is to provide an alternative means
 * of monitoring 128 digital inputs and controlling 128 digital
 * outputs via 8-bit event "opcodes" delivered to the user either
 * via a serial UART (async_serial_evgpio) or single 16-bit WISHBONE
 * register (wb_evgpio).
 *
 * The most common mechanism for monitoring for changes of state
 * among 128 inputs involves having software poll lots of 16-bit
 * registers and use boolean bitwise operations to query interested
 * pin states.  An EVGPIO core is instead constantly monitoring
 * inputs for posedges or negedges and will deliver an 8-bit opcode
 * of the following format when a delta from the previously sent
 * state is detected.
 *
 *   bit 7: posedge detect (1- low to high event, 0- high to low event) 
 *   bits 6-0: input number (0-127)
 *
 * The wb_evgpio core will assert an IRQ when an event is pending
 * whereas the async_serial_evgpio core transmits a 8N1 byte at the
 * chosen baudrate.  On the wb_evgpio core, software is then expected
 * to read 8-bit opcodes until the IRQ is deasserted.  The IRQ is
 * also reported in bit 8 of the 16-bit read register as a convenience
 * for the IRQ handler routine in knowing when the input states are
 * stable and no more event bytes are waiting to be read.
 *
 * By default, high numbered input bit events are lower priority
 * than low order bits.  This means that if there are multiple pins
 * changing state simultaneously, the low order bits' events will
 * always be reported first.  Keep in mind that if there are more
 * events happening on the inputs than there is bus or serial bandwidth
 * to support reporting, the wb_evgpio IRQ may never deassert and
 * the async_serial_evgpio core may never stop transmitting.  When
 * possible, software layers should make use of userspace IRQs and
 * process priorities or other means to control maximum acceptable
 * CPU utilization.  A naively written kernel mode driver could
 * simply lock up the CPU if somebody sent a 100Mhz square wave on
 * a monitored input pin.  Using the async_serial_evgpio core however
 * will not likely need such consideration if connected to a proper
 * UART such as a TS-XUART.
 *
 * To manipulate output bits, the same opcode format is used.
 * Manipulating output bits via an EVGPIO core avoids complications
 * from multiple processes using read-modify-write register bus cycles
 * and reduces the needs for customers to understand boolean bitwise 
 * operaters. (A good thing for non-embedded programmers!)
 *
 * When more than 64 pins are used for input, it may be difficult to
 * meet timing.  If so, contact Jesse as he will have to add another
 * clock period for allowing the arbiter 2 clocks instead of 1 and add
 * a multicycle timing constraint.
 */

module wb_evgpio(
  input wb_clk_i, wb_rst_i, wb_cyc_i, wb_stb_i, wb_we_i, 
  input [31:0] wb_adr_i,
  input [15:0] wb_dat_i, 
  output reg [15:0] wb_dat_o = 16'd0, 
  output reg wb_ack_o = 1'b0,

  input [127:0] gpios_i,
  output reg [127:0] gpios_o = 128'd0,
  output irq_o
);
parameter number_of_gpios = 128;

wire [127:0] gpio_en, code_valid;
wire [7:0] codes[0:127], code_o[0:127];
reg [1:0] state = 2'd0;
wire pending;
assign irq_o = pending;
genvar x;
generate
for (x = 0; x < 128; x = x + 1) begin: gpios
  gpio_scanner gpio_scanner_core(
    .clk_i(wb_clk_i),
    .rst_i(wb_rst_i),
    .enable_i((gpio_en[x] && state == 2'd2 && !wb_ack_o && wb_cyc_i && wb_stb_i && !wb_we_i && pending) || (state == 2'd0)),
    .gpio_i(x >= number_of_gpios ? 1'b0 : gpios_i[x]),
    .code_valid_o(code_valid[x]),
    .code_o(code_o[x]),
    .negedge_code_i({1'b0, x[6:0]}),
    .posedge_code_i({1'b1, x[6:0]})
  );
  assign codes[x] = gpio_en[x] ? code_o[x] : 8'd0;
end
endgenerate

integer i;
reg [7:0] code;
always @(*) begin
  code = 8'd0;
  for (i = 0; i < 128; i = i + 1) code = code | codes[i];
end

ts_arbiter code_arbiter(
  .clk_i(wb_clk_i),
  .rst_i(wb_rst_i),
  .req_i(code_valid),
  .grant_o(gpio_en),
  .pending_o(pending),
  .ack_i(state == 2'd3 ? 1'b1 : 1'b0)
);

reg [15:0] wb_dat_i_q = 16'd0;
reg [2:0] wb_adr_i_minus1_q = 3'd0;
reg wb_wr_16 = 1'b0, wb_wr = 1'b0;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    wb_wr <= 1'b0;
    wb_wr_16 <= 1'b0;
    wb_ack_o <= 1'b0;
    wb_dat_i_q <= 16'd0;
    wb_adr_i_minus1_q <= 3'd0;
    state <= 2'd0; 
    gpios_o <= 128'd0;
    wb_dat_o <= 16'd0;
  end else begin
    wb_wr <= 1'b0;
    wb_wr_16 <= 1'b0;
    wb_ack_o <= 1'b0;
    wb_dat_i_q <= wb_dat_i;
    wb_adr_i_minus1_q <= wb_adr_i[3:0] - 1'b1;
    if (wb_we_i && wb_cyc_i && wb_stb_i && !wb_ack_o) begin
      if (wb_adr_i > 32'd0) wb_wr_16 <= 1'b1;
      else wb_wr <= 1'b1;
      wb_ack_o <= 1'b1;
    end
    if (wb_wr) gpios_o[wb_dat_i_q[6:0]] <= wb_dat_i_q[7];
    else if (wb_wr_16) gpios_o[wb_adr_i_minus1_q*16+:16] <= wb_dat_i_q[15:0];
  
    case (state)
    2'd0: state <= 2'd1; /* Waiting for a key */
    2'd1: state <= 2'd2; /* Arbitrate code */
    2'd2: begin
      if (!pending) state <= 2'd0; 
      if (wb_cyc_i && wb_stb_i && !wb_we_i && !wb_ack_o && pending) begin
        state <= 2'd3;
        wb_ack_o <= 1'b1;
        wb_dat_o <= {8'd1, code};
      end
      if (wb_cyc_i && wb_stb_i && !wb_we_i && !wb_ack_o && !pending) begin
        wb_ack_o <= 1'b1;
        wb_dat_o <= {8'd0, code};
      end
    end
    2'd3: state <= 2'd2;
    endcase
  end
end
endmodule


module async_serial_evgpio(
  input clk_i, rst_i, rx_i, cts_i, /* when cts_i is 1, suspends TX output */
  input tx_baud_i, /* 1x baud rate (clk-enable, width = 1 clk_i period) */
  input rx_baud_i, /* 8x baud rate (clk-enable, width = 1 clk_i period) */
  output tx_o, txen_o,
  input [127:0] gpios_i,
  output reg [127:0] gpios_o = 128'd0
);
parameter clk_freq = 75000000;
parameter baud = 115200;
parameter internal_baudgen = 1'b1;
parameter number_of_gpios = 128;
parameter debounce = 1'b0; /* Make sure you turn this off for non-keypads! */

integer clks_per_baud = (1.0 / baud) / (1.0 / clk_freq);
integer clks_per_baud_8x = ((1.0 / baud) / (1.0 / clk_freq)) / 8;
reg [31:0] clks = 32'd0, clks2 = 32'd0;
reg int_baud = 1'b0, int_baud_8x = 1'b0;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    int_baud <= 1'b0;
    int_baud_8x <= 1'b0;
    clks <= 32'd0;
    clks2 <= 32'd0;
  end else begin
    int_baud <= 1'b0;
    int_baud_8x <= 1'b0;
    if (clks > (clks_per_baud - 1)) begin
      clks <= 32'd0;
      clks2 <= 32'd0;
      int_baud <= 1'b1;
      int_baud_8x <= 1'b1;
    end else begin
      clks <= clks + 1'b1;
      clks2 <= clks2 + 1'b1;
    end
    if (clks2 > (clks_per_baud_8x - 1)) begin
      int_baud_8x <= 1'b1;
      clks2 <= 32'd0;
    end
  end
end

wire [127:0] gpio_en, code_valid, gpios_debounced[0:127];
wire [7:0] codes[0:127], code_o[0:127];
reg [1:0] state = 2'd0;
wire tbre, pending;
genvar x;
generate
for (x = 0; x < 128; x = x + 1) begin: gpios
  gpio_debouncer #(.clk_freq(clk_freq)) gpio_debouncer_core (
    .clk_i(clk_i),
    .rst_i(rst_i),
    .gpio_i({127'd0, gpios_i[x]}),
    .gpio_o(gpios_debounced[x])
  );
  gpio_scanner gpio_scanner_core(
    .clk_i(clk_i),
    .rst_i(rst_i),
    .enable_i((gpio_en[x] && state == 2'd2 && tbre) || (state == 2'd0)),
    .gpio_i(x >= number_of_gpios ? 1'b0 : 
      (debounce ? gpios_debounced[x][0] : gpios_i[x])),
    .code_valid_o(code_valid[x]),
    .code_o(code_o[x]),
    .negedge_code_i({1'b0, x[6:0]}),
    .posedge_code_i({1'b1, x[6:0]})
  );
  assign codes[x] = gpio_en[x] ? code_o[x] : 8'd0;
end
endgenerate

integer i;
reg [7:0] code;
always @(*) begin
  code = 8'd0;
  for (i = 0; i < 128; i = i + 1) code = code | codes[i];
end

ts_arbiter code_arbiter(
  .clk_i(clk_i),
  .rst_i(rst_i),
  .req_i(code_valid),
  .grant_o(gpio_en),
  .pending_o(pending),
  .ack_i(state == 2'd3 ? 1'b1 : 1'b0)
);

ts_async_serial_tx txcore(
  .clk_i(clk_i),
  .rst_i(rst_i),
  .enable_i(internal_baudgen ? int_baud : tx_baud_i),
  .tload_i(tbre && pending),
  .tbre_o(tbre),
  .d_i(code),
  .tx_o(tx_o),
  .cts_i(cts_i),
  .tactive_o(txen_o)
);

wire dr;
wire [7:0] rxdata;
ts_async_serial_rx rxcore(
  .clk_i(clk_i),
  .rst_i(rst_i),
  .enable_i(internal_baudgen ? int_baud_8x : rx_baud_i),
  .rx_i(rx_i),
  .rd_i(dr),
  .dr_o(dr),
  .d_o(rxdata)
);

always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    gpios_o <= 128'd0;
    state <= 2'd0;
  end else begin 
    if (dr) gpios_o[rxdata[6:0]] <= rxdata[7];
    case (state)
    2'd0: state <= 2'd1; /* Waiting for a key */
    2'd1: state <= 2'd2; /* Arbitrate code */
    2'd2: if (!pending) state <= 2'd0; else if (tbre) state <= 2'd3;
    2'd3: state <= 2'd2;
    endcase
  end
end
endmodule


module ts_arbiter( /* Priority arbiters are smaller/faster than roundrobin */
  input clk_i,
  input rst_i,
  input [127:0] req_i,
  output reg [127:0] grant_o = 128'd0,
  output reg [6:0] select_o = 7'd0,
  output reg pending_o = 1'b0,
  input ack_i
);
parameter roundrobin_opt = 1'b0; /* else a priority arb, LSBs = highest prio */

wire [127:0] req_rr_tmp = req_i & ~((grant_o - 1) | grant_o);
wire [127:0] req_rr = |req_rr_tmp ? req_rr_tmp : req_i;
wire [127:0] next_grant_rr = req_rr & (~req_rr + 1);
wire [127:0] next_grant_prio = req_i & (~req_i + 1);
wire [127:0] next_grant = roundrobin_opt ? next_grant_rr : next_grant_prio;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    grant_o <= 128'd0;
    pending_o <= 1'b0;
  end else begin
    if ((pending_o && ack_i) || !pending_o) begin
      grant_o <= next_grant;
      pending_o <= |next_grant;
    end
  end
end

integer i;
always @(*) begin
  select_o = 7'd0;
  for (i = 0; i < 128; i = i + 1) if (grant_o[i]) select_o = select_o | i;
end
endmodule


module gpio_debouncer(
  input clk_i, rst_i,
  input [127:0] gpio_i,
  output reg [127:0] gpio_o = 128'd0
);
parameter clk_freq = 75000000;
parameter min_stable_ms = 20;

integer stable_cnt_min = (min_stable_ms * 1e-3) / (1.0 / clk_freq); 
integer stable_cnt_bit, i;
always @(*) begin /* Effectively rounds up to the nearest power of two */
  stable_cnt_bit = 0;
  for (i = 0; i < 32; i = i + 1) if (stable_cnt_min[i]) stable_cnt_bit = i + 1;
end

reg [31:0] cnt = 32'd0, basecnt = 32'd0; /* basecnt resource is shareable */
reg [127:0] gpio_q = 128'd0, gpio_qq = 128'd0;
reg stable = 1'b0;
reg [1:0] basecnt_sr;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    basecnt <= 32'd0;
    cnt <= 32'd0;
    gpio_q <= 128'd0;
    gpio_qq <= 128'd0;
    stable <= 1'b0;
    basecnt_sr <= 2'd0;
    gpio_o <= 128'd0;
  end else begin
    basecnt <= basecnt + 1'b1;
    basecnt_sr <= {basecnt_sr[0], basecnt[stable_cnt_bit - 3]};
    if (^basecnt_sr) begin
      if (stable) cnt <= cnt + 1'b1;
      if (cnt[3]) begin
        gpio_o <= gpio_q;
        cnt[3] <= 1'b0;
      end
    end
    gpio_q <= gpio_i;
    gpio_qq <= gpio_q;
    if (gpio_q === gpio_qq) stable <= 1'b1; else stable <= 1'b0;
    if (!stable) cnt <= 32'd0;
  end
end
endmodule


module gpio_scanner(
  input clk_i, rst_i, enable_i, gpio_i,
  output code_valid_o,
  output [7:0] code_o,
  input [7:0] posedge_code_i, negedge_code_i
);

assign code_valid_o = (gpio_q && !gpio_qq) || (!gpio_q && gpio_qq);
assign code_o = gpio_q ? posedge_code_i : negedge_code_i;
reg gpio_q = 1'b0, gpio_qq = 1'b0;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    gpio_q <= 1'b0;
    gpio_qq <= 1'b0;
  end else if (enable_i) begin
    gpio_q <= gpio_i;
    gpio_qq <= gpio_q;
  end
end
endmodule


module ts_async_serial_tx(
  input clk_i, rst_i, tload_i, enable_i, /* one enable per bit time */
  input [7:0] d_i, 
  input cts_i, /* when 0, tx output enabled */
  output reg tx_o = 1'b1, /* async serial TX data out */
  output reg tre_o = 1'b1, /* transmitter empty */
  output reg tbre_o = 1'b1, /* transmitter buffer empty */
  output reg tactive_o = 1'b0
);

reg [7:0] tbr = 8'd0, tr = 8'hff;
reg [3:0] cbit = 4'd0;
reg cts_i_q = 1'b0;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    tbr <= 8'd0;
    tr <= 8'hff;
    cbit <= 4'd0;
    cts_i_q <= 1'b0;
    tx_o <= 1'b1;
    tre_o <= 1'b1;
    tbre_o <= 1'b1;
    tactive_o <= 1'b0;
  end else begin
    cts_i_q <= cts_i;
    if (enable_i) begin
      tactive_o <= !tre_o;
      cbit <= cbit + 1'b1;
      tr <= {1'b1, tr[7:1]};
      tx_o <= tr[0];
      if (cbit[3]) tre_o <= 1'b1;
      if (!tbre_o && tre_o && !cts_i_q) begin
        tactive_o <= 1'b1;
        tr <= tbr;
        tre_o <= 1'b0;
        tbre_o <= 1'b1;
        tx_o <= 1'b0;
        cbit <= 4'd0;
      end
    end
    if (tload_i) begin
      tbr <= d_i[7:0];
      tbre_o <= 1'b0;
    end
  end
end
endmodule


module ts_async_serial_rx(
  input clk_i, rst_i, enable_i, rx_i, rd_i,
  output reg dr_o = 1'b0,
  output reg [7:0] d_o = 8'd0,
  output reg oerr_o = 1'b0
);

reg start = 1'b0;
reg [3:0] cbit = 4'd0;
reg [2:0] csam = 3'd0;
reg [7:0] sr = 8'd0;

always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    dr_o <= 1'b0;
    d_o <= 8'd0;
    oerr_o <= 1'b0;
    start <= 1'b0;
    cbit <= 4'd0;
    csam <= 3'd0;
    sr <= 8'd0;
  end else begin
    if (rd_i) dr_o <= 1'b0;
  
    if (enable_i) begin
      csam <= csam + 1'b1;
      if (csam == 3'd3) sr <= {rx_i, sr[7:1]};
      if (!start) begin
        csam <= 3'd0;
        cbit <= 4'd15;
        if (!rx_i) start <= 1'b1;
      end else if (csam == 3'd3) begin
        cbit <= cbit + 1'b1;
        if (cbit[3:2] == 2'b10) begin // stop bit
          // On framing errors we just ignore the byte and immediately
          // start looking again for the next start bit.
          if (rx_i) begin
            if (dr_o) oerr_o <= 1'b1;
            else oerr_o <= 1'b0;
            dr_o <= 1'b1;
            d_o <= sr;
          end
          start <= 1'b0;
        end else if (cbit[3] == 1'b1) begin   // start bit
          // We must double-check our start bits.
          if (rx_i == 1'b1) begin
            start <= 1'b0;
          end
        end
      end
    end
  end
end
endmodule

`ifdef TESTBENCH_SERIAL
module test();

reg clk, baud;
reg [127:0] gpios;
wire [127:0] gpios_o;
wire rxtx;
async_serial_evgpio #(.clk_freq(100000000)) uut_tx(
  .clk_i(clk),
  .gpios_i(gpios),
  .tx_o(rxtx),
  .cts_i(1'b0)
);
async_serial_evgpio #(.clk_freq(100000000)) uut_rx(
  .clk_i(clk),
  .gpios_o(gpios_o),
  .rx_i(rxtx)
);

always #5 clk = !clk;

initial begin
/*
  $dumpfile("dump.vcd");
  $dumpvars(3, test);
*/
  $monitor("%t gpios_o: %x", $time, gpios_o);
  clk = 0;
  baud = 0;
  gpios = 0;
  #10;
  gpios = 128'hdeadbeef;
  #2100000;
  $finish;
end
endmodule
`endif

`ifdef TESTBENCH_WB
module test();

reg clk, rst, cyc, stb, cyc2, stb2;
wire ack, ack2, irq;
wire [15:0] dat;
reg [15:0] dat2;
reg [127:0] gpios;
wire [127:0] gpios_o;
wb_evgpio uut_rx(
  .wb_clk_i(clk),
  .wb_rst_i(rst),
  .wb_cyc_i(cyc),
  .wb_stb_i(stb),
  .wb_dat_o(dat),
  .wb_we_i(1'b0),
  .wb_ack_o(ack),
  .gpios_i(gpios),
  .irq_o(irq)
);
wb_evgpio uut_tx(
  .wb_clk_i(clk),
  .wb_rst_i(rst),
  .wb_cyc_i(cyc2),
  .wb_stb_i(stb2),
  .wb_we_i(1'b1),
  .wb_dat_i(dat2),
  .wb_ack_o(ack2),
  .gpios_o(gpios_o)
);

always #5 clk = !clk;

always @(posedge clk or posedge rst) begin
  if (rst) begin
    cyc <= 1'b0;
    stb <= 1'b0;
    cyc2 <= 1'b0;
    stb2 <= 1'b0;
    dat2 <= 16'd0;
  end else begin
    if (irq && !cyc && !cyc2) begin
      $display("%t got irq", $time);
      cyc <= 1'b1;
      stb <= 1'b1;
    end
    if (cyc && ack) begin
      $display("%t got dat: %x", $time, dat);
      dat2 <= dat;
      cyc <= 1'b0;
      stb <= 1'b0;
      cyc2 <= 1'b1;
      stb2 <= 1'b1;
    end
    if (cyc2 && ack2) begin
      $display("%t sent dat: %x", $time, dat2);
      cyc2 <= 1'b0;
      stb2 <= 1'b0;
    end
  end
end

initial begin
/*
  $dumpfile("dump.vcd");
  $dumpvars(3, test);
*/
  rst = 1;
  #10;
  rst = 0;
  #10;
  $monitor("%t gpios_o: %x", $time, gpios_o);
  clk = 0;
  gpios = 0;
  #10;
  gpios = 128'hffffffff;
  #1000;
  $display("Switching!");
  gpios = 128'h8000ffff;
  #10000;
  $display("Switching 2!");
  gpios = 128'h0000ffff;
  #50;
  gpios = 128'h8000ffff;
  #1000;
  $finish;
end
endmodule
`endif
